// Read/Write request processors for sdla driver

#include "stddcls.h"
#include "driver.h"
#include "sdlapci.h"


///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS DispatchCreate(PDEVICE_OBJECT fdo, PIRP Irp)
{							// DispatchCreate
	PAGED_CODE();
	KdPrint((DRIVERNAME " - IRP_MJ_CREATE\n"));
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
	
	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
	
	// Claim the remove lock in Win2K so that removal waits until the
	// handle closes. Don't do this in Win98, however, because this
	// device might be removed by surprise with handles open, whereupon
	// we'll deadlock in HandleRemoveDevice waiting for a close that
	// can never happen because we can't run the user-mode code that
	// would do the close.
	
	NTSTATUS status;
	if (win98)
		status = STATUS_SUCCESS;
	else 
		status = IoAcquireRemoveLock(&pdx->RemoveLock, stack->FileObject);
	
	if (NT_SUCCESS(status))
	{						// okay to open
		if (InterlockedIncrement(&pdx->handles) == 1)
		{					// first open handle
		}					// okay to open
	}					// first open handle
	return CompleteRequest(Irp, status, 0);
}							// DispatchCreate

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS DispatchClose(PDEVICE_OBJECT fdo, PIRP Irp)
{							// DispatchClose
	PAGED_CODE();
	KdPrint((DRIVERNAME " - IRP_MJ_CLOSE\n"));
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
	if (InterlockedDecrement(&pdx->handles) == 0)
	{						// no more open handles
	}						// no more open handles
	
	// Release the remove lock to match the acquisition done in DispatchCreate
	
	if (!win98)
		IoReleaseRemoveLock(&pdx->RemoveLock, stack->FileObject);
	
	return CompleteRequest(Irp, STATUS_SUCCESS, 0);
}							// DispatchClose

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS StartDevice(PDEVICE_OBJECT fdo, PCM_PARTIAL_RESOURCE_LIST raw, PCM_PARTIAL_RESOURCE_LIST translated)
{							// StartDevice
	PDEVICE_EXTENSION pdx;// = (PDEVICE_EXTENSION) fdo->DeviceExtension;
	NTSTATUS status;
	
	static LONG device_count = 0;

	// Identify the I/O resources we're supposed to use.
	PHYSICAL_ADDRESS membase, membase1, membase2;
	BOOLEAN gotmemory = FALSE;
	
	PHYSICAL_ADDRESS CtrlRegisterAddress; // 8-byte physical addresses.
	ULONG CtrlRegisterLengInBytes;

	ULONG BASE_REGISTER_ADDRESS_RESOURCE_NUMBER;
	ULONG BASE_MEMORY_ADDRESS_CPU_A_RESOURCE_NUMBER;
	ULONG BASE_MEMORY_ADDRESS_CPU_B_RESOURCE_NUMBER;
	ULONG S514_2CPU_RES_COUNT;

	if(win98)
	{
		BASE_REGISTER_ADDRESS_RESOURCE_NUMBER = BASE_REGISTER_ADDRESS_RESOURCE_NUMBER_WIN_98;
		BASE_MEMORY_ADDRESS_CPU_A_RESOURCE_NUMBER = BASE_MEMORY_ADDRESS_CPU_A_RESOURCE_NUMBER_WIN_98;
		BASE_MEMORY_ADDRESS_CPU_B_RESOURCE_NUMBER = BASE_MEMORY_ADDRESS_CPU_B_RESOURCE_NUMBER_WIN_98;
		S514_2CPU_RES_COUNT = S514_2CPU_RES_COUNT_WIN_98;
	}else
	{
		BASE_REGISTER_ADDRESS_RESOURCE_NUMBER = BASE_REGISTER_ADDRESS_RESOURCE_NUMBER_WIN_2000;
		BASE_MEMORY_ADDRESS_CPU_A_RESOURCE_NUMBER = BASE_MEMORY_ADDRESS_CPU_A_RESOURCE_NUMBER_WIN_2000;
		BASE_MEMORY_ADDRESS_CPU_B_RESOURCE_NUMBER = BASE_MEMORY_ADDRESS_CPU_B_RESOURCE_NUMBER_WIN_2000;
		S514_2CPU_RES_COUNT = S514_2CPU_RES_COUNT_WIN_2000;
	}

	if (!translated)
		return STATUS_DEVICE_CONFIGURATION_ERROR;		// no resources assigned??
	
	PCM_PARTIAL_RESOURCE_DESCRIPTOR resource = translated->PartialDescriptors;
	ULONG nres = translated->Count;
	
	KdPrint((DRIVERNAME " - translated->Count %d\n", translated->Count));
	//pdx->Count = translated->Count;
	
	for (ULONG i = 0; i < nres; ++i, ++resource)
	{						// for each resource
		switch (resource->Type)
		{					// switch on resource type
		case CmResourceTypeMemory:
			
			KdPrint((DRIVERNAME " - i : %d\n", i));
			
			if(i == BASE_REGISTER_ADDRESS_RESOURCE_NUMBER)//1 for win98, 0 for win2000
			{	membase = resource->u.Memory.Start;
				gotmemory = TRUE;
			}else if(i == BASE_MEMORY_ADDRESS_CPU_A_RESOURCE_NUMBER){//2 for win98 and win2000
				membase1 = resource->u.Memory.Start;
				gotmemory = TRUE;
			}else if(i == BASE_MEMORY_ADDRESS_CPU_B_RESOURCE_NUMBER){//3 for win98 and 4 for win2000
				membase2 = resource->u.Memory.Start;
				gotmemory = TRUE;
			}else{
				;//do nothing
			}

			break;

		default:
			KdPrint((DRIVERNAME " - Unexpected I/O resource type %d\n", resource->Type));
			break;
		}					// switch on resource type
	}						// for each resource
	
	pdx = (PDEVICE_EXTENSION)ExAllocatePool(NonPagedPool, sizeof(DEVICE_EXTENSION));
	RtlZeroMemory(pdx, sizeof(DEVICE_EXTENSION));
	
	// Verify that we got all the resources we were expecting
	if (!(TRUE && gotmemory	))
	{
		KdPrint((DRIVERNAME " - Didn't get expected I/O resources\n"));
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}
	
	pdx->BaseRegisterAddress = (PUCHAR)MmMapIoSpace(membase, 0x100, MmNonCached);
	if (!pdx->BaseRegisterAddress)
	{
		KdPrint((DRIVERNAME " - Unable to map BaseRegisterAddress %I64X, length %X\n", 
			pdx->BaseRegisterAddress, 0x100));
		return STATUS_INSUFFICIENT_RESOURCES;
	}			
	
	pdx->BaseMemoryAddress = (PUCHAR)MmMapIoSpace(membase1, 0x100000, MmNonCached);
	if (!pdx->BaseMemoryAddress)
	{
		KdPrint((DRIVERNAME " - Unable to map BaseMemoryAddress %I64X, length %X\n", 
			pdx->BaseMemoryAddress, 0x100000));
		return STATUS_INSUFFICIENT_RESOURCES;
	}			
	
	pdx->Memory_S514_CTRL_REG_BYTE = pdx->BaseMemoryAddress + S514_CTRL_REG_BYTE;
	pdx->CPU = CPU_A; //cpu A
	
	//Copy pointers to each device extension into devices array
	//because each device can be used by number of logical cards.
	for(i = 0; i < MAX_CARDS; i++)
	{
		if(device_count == sdla[i])
		{	dev_array[i] = pdx;
#ifdef DEBG
		DbgPrint("sdla[%d] will use device:%d\n", i, device_count);
#endif
		}
	}
	
	detected_devices_array[detected_devices_count] = pdx;
	detected_devices_count++;
	device_count++;

	//if two CPUs, create another device
	if(translated->Count == S514_2CPU_RES_COUNT)
	{	
		PDEVICE_EXTENSION pdx_cpu_B = (PDEVICE_EXTENSION)ExAllocatePool(NonPagedPool, sizeof(DEVICE_EXTENSION));
		RtlZeroMemory(pdx_cpu_B, sizeof(DEVICE_EXTENSION));

#ifdef DEBG
		DbgPrint("Detected 2 CPUs\n");
#endif

		pdx_cpu_B->BaseRegisterAddress = pdx->BaseRegisterAddress;

		pdx_cpu_B->BaseMemoryAddress = (PUCHAR)MmMapIoSpace(membase2, 0x100000, MmNonCached);
		if (!pdx_cpu_B->BaseMemoryAddress)
		{
			KdPrint((DRIVERNAME " - Unable to map BaseMemoryAddress %I64X, length %X\n", 
				pdx_cpu_B->BaseMemoryAddress, 0x100000));
			return STATUS_INSUFFICIENT_RESOURCES;
		}			
		
		pdx_cpu_B->Memory_S514_CTRL_REG_BYTE = pdx_cpu_B->BaseMemoryAddress + S514_CTRL_REG_BYTE;
		pdx_cpu_B->CPU = CPU_B; //cpu B
		
		//Copy pointers to each device extension into devices array
		//because each device can be used by number of logical cards.
		for(i = 0; i < MAX_CARDS; i++)
		{
			if(device_count == sdla[i])
			{	dev_array[i] = pdx_cpu_B;
#ifdef DEBG
			DbgPrint("sdla[%d] will use device:%d\n", i, device_count);
#endif
			}
		}
		
		detected_devices_array[detected_devices_count] = pdx_cpu_B;
		detected_devices_count++;
		device_count++;
	}

	return STATUS_SUCCESS;
}							// StartDevice

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

VOID StopDevice(IN PDEVICE_OBJECT fdo, BOOLEAN oktouch /* = FALSE */)
{							// StopDevice
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;

#ifdef DEBG
	DbgPrint("StopDevice()\n");
#endif
	
	if (pdx->membase)
		MmUnmapIoSpace(pdx->membase, pdx->memsize);
	pdx->membase = NULL;
}							// StopDevice
